/*
 * GUTS - Generic Utilities 
 * Copyright (c) 2006-2010 Gino van den Bergen, DTECTA 
 * http://www.dtecta.com
 *
 * Permission to use, copy, modify, distribute and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear
 * in supporting documentation.  Gino van den Bergen makes no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied warranty.
 */

#ifndef GUTS_ARRAY_HPP
#define GUTS_ARRAY_HPP

#include <memory>
#include <algorithm>

#include <consolid.h>


namespace guts
{
    template <typename Data, typename Alloc = std::allocator<Data> >
    class Array 
    {
    public:
        typedef Data value_type;
        typedef Data* iterator;
        typedef const Data* const_iterator;

        Array() 
            : m_count(0)
            , m_data(0) 
        {}

        explicit Array(size_t count)
            : m_count(count)
            , m_data(count != 0 ? Alloc().allocate(count) : 0) 
        {}
        
        Array(size_t count, const Data* data) 
            : m_count(count)
            , m_data(count != 0 ? Alloc().allocate(count) : 0) 
        {
            ASSERT(count == 0 || data != 0);
            std::copy(data, data + count, m_data);
        }
        
        Array(const Array<Data>& rhs)
            : m_count(rhs.size())
            , m_data(rhs.size() != 0 ? Alloc().allocate(rhs.size()) : 0)
        {
            std::copy(rhs.begin(), rhs.end(), m_data);
        }

        ~Array() 
        { 
            Alloc().deallocate(m_data, m_count); 
        } 
        
        Array<Data>& operator=(const Array<Data>& rhs)
        {
            if (this != &rhs)
            {
                Data* data = rhs.size() != 0 ? Alloc().allocate(rhs.size()) : 0;
                std::copy(rhs.begin(), rhs.end(), data);
                Alloc().deallocate(m_data, m_count); 
                m_data = data;
                m_count = rhs.size();
            }
            return *this;
        }
        
        const Data& operator[](size_t i) const
        {
            ASSERT(i < m_count);
            return m_data[i];
        }
        
        Data& operator[](size_t i)
        { 
            ASSERT(i < m_count);
            return m_data[i]; 
        }

        size_t size() const
        {
            return m_count; 
        }

        bool empty() const
        {
            return m_count == 0;
        }

        const Data* begin() const
        {
            return m_data;
        }
        
        Data* begin()
        {
            return m_data;
        }
        
        const Data* end() const
        {
            return m_data + m_count;
        }
        
        Data* end()
        {
            return m_data + m_count;
        }
        
    private:
        size_t m_count;
        Data*  m_data;
    };
}
  
#endif

