/*
 * GUTS - Generic Utilities 
 * Copyright (c) 2006, Gino van den Bergen, DTECTA 
 * http://www.dtecta.com
 *
 * Permission to use, copy, modify, distribute and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear
 * in supporting documentation.  Gino van den Bergen makes no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied warranty.
 */

#ifndef GUTS_BINOMIALQUEUE_HPP
#define GUTS_BINOMIALQUEUE_HPP

#include <vector>

#include <consolid/consolid.h>

#include "Invariant.hpp"

namespace guts
{
    template <typename Key>
    class BinomialQueue
    {
    public:   
        BinomialQueue();
                 
        bool empty() const;
        size_t push(const Key& value);
        void decrease_key(size_t index, Key value);
        void destroy(size_t index);
        Key pop();
 
    private:
        static const size_t NO_REF = ~0;   
        static const size_t PARENT_REF = 1 << 31;      

        struct Node
        {
            Node(const Key& value)
                : m_value(value)
                , m_rchild(NO_REF)
                , m_lchild(NO_REF)
            {}

            Key     m_value;
            size_t m_rchild;
            size_t m_lchild;
        };  
        
        typedef std::vector<Node> NodeStore_t;

           
        void binomial(size_t index, int height) const;

        size_t link(size_t a, size_t b);
        size_t orphan(size_t index, size_t prev);
        void reunite(size_t child); 
        void insert(size_t index);
        void remove(size_t index);
        size_t remove_sub(size_t index, size_t prev);
    
        
        size_t      m_trees[32];
        int         m_height;
        NodeStore_t m_nodes;
        size_t      m_free; 

        INVARIANT
        {
            for (int i = 0; i != m_height; ++i)
            {   
                if (m_trees[i] != NO_REF)
                {
                    size_t index = m_trees[i];
                    ASSERT(m_nodes[index].m_rchild == NO_REF);
                    binomial(index, i);
                }
            }
            ASSERT(m_height == 0 || m_trees[m_height - 1] != NO_REF);
        }
    }; 
    
    
    template <typename Key>
    BinomialQueue<Key>::BinomialQueue()
        : m_height(0)
        , m_free(NO_REF)
    {
        GUARD_INVARIANT_AT_EXIT(this);

        for (size_t i = 0 ; i != 32; ++i)
		{
			m_trees[i] = NO_REF;
		}
    }  
            
    template <typename Key>
    inline
    bool BinomialQueue<Key>::empty() const
    {
        return m_height == 0;
    }

    template <typename Key>
    size_t BinomialQueue<Key>::push(const Key& value)
    {
        size_t index;
        
        if (m_free == NO_REF)
        {
            index = size_t(m_nodes.size());
            m_nodes.push_back(Node(value));
        }
        else
        {
            index = m_free;
            m_free = m_nodes[index].m_rchild;
            new (&m_nodes[index]) Node(value);
        }

        insert(index);

        return index;
    }

        
    template <typename Key>
    void BinomialQueue<Key>::decrease_key(size_t index, Key value)
    {
        remove(index); 
        
        new (&m_nodes[index]) Node(value);

        insert(index);
    }
    
    template <typename Key>
    void BinomialQueue<Key>::destroy(size_t index)
    {
        remove(index); 
        
        m_nodes[index].m_rchild = m_free;
        m_free = index;
    }
        
    template <typename Key>
    Key BinomialQueue<Key>::pop()
    {
        ASSERT(!empty());
        
        int i = 0;
        while (m_trees[i] == NO_REF)
        {
            ++i;
        }
        
        size_t best = i;
        Key result = m_nodes[m_trees[i]].m_value;

        while (++i != m_height)    
        {
            if (m_trees[i] != NO_REF && m_nodes[m_trees[i]].m_value < result)
            {
                best = i;
                result = m_nodes[m_trees[i]].m_value;
            }
        }
        
        size_t index = m_trees[best];
        size_t child = orphan(index, NO_REF);
            
        reunite(child);
        
        m_nodes[index].m_rchild = m_free;
        m_free = index;

        return result;
    }    

    template <typename Key>
    void BinomialQueue<Key>::binomial(size_t index, int height) const 
    {
        ASSERT(index != NO_REF);

        size_t child = m_nodes[index].m_lchild;

        if (height == 0)
        {
            ASSERT(child == NO_REF);
        }
        else
        {
            do
            {
                --height;
                ASSERT(m_nodes[index].m_value <= m_nodes[child].m_value);
                binomial(child, height);
                child = m_nodes[child].m_rchild;
            }
            while (height != 0);

            ASSERT(child == (PARENT_REF|index));
        }
    }
       
    template <typename Key>
    inline
    size_t BinomialQueue<Key>::link(size_t a, size_t b)
    {
        if (m_nodes[b].m_value < m_nodes[a].m_value)
        {
            std::swap(a, b);
        }
            
        m_nodes[b].m_rchild = m_nodes[a].m_lchild != NO_REF ? m_nodes[a].m_lchild : PARENT_REF | a;
        m_nodes[a].m_lchild = b; 
        m_nodes[a].m_rchild = NO_REF; 
        
        return a;
    }

    template <typename Key>
    inline
    size_t BinomialQueue<Key>::orphan(size_t index, size_t prev)
    {
        size_t child = m_nodes[index].m_lchild;
        while ((child & PARENT_REF) == 0)
        {
            size_t next = m_nodes[child].m_rchild;
            m_nodes[child].m_rchild = prev;
            prev = child;
            child = next;
        } 
        return prev;
    }
        
    template <typename Key>
    inline
    void BinomialQueue<Key>::reunite(size_t child)
    {
        int i = 0;
        while (child != NO_REF && m_trees[i] == NO_REF)
        { 
            m_trees[i] = child;
            child = m_nodes[child].m_rchild;
            m_nodes[m_trees[i]].m_rchild = NO_REF;
            ++i;
        } 
        
        if (child != NO_REF)
        {
            size_t next = m_nodes[child].m_rchild;
            size_t carry = link(m_trees[i], child);
            m_trees[i] = NO_REF;
            child = next;
            ++i;
            
            while (child != NO_REF)
            {
                next = m_nodes[child].m_rchild;
                carry = link(carry, child);
                child = next;
                ++i;
            }

            m_trees[i] = carry;
        }
        else
        {
            m_trees[i] = NO_REF;
            
            if (i == m_height-1)
            {
                --m_height;
            }
        }
    }

    template <typename Key>
    inline
    void BinomialQueue<Key>::insert(size_t index)
    {
        GUARD_INVARIANT(this);

        int i = 0;
        while (m_trees[i] != NO_REF)
        {
            index = link(m_trees[i], index); 
            m_trees[i] = NO_REF;
            ++i;
        }
        m_trees[i] = index;

        if (i == m_height)
        {
            ++m_height;
        }

    }
        
    template <typename Key>
    inline
    void BinomialQueue<Key>::remove(size_t index)
    {
        GUARD_INVARIANT(this);
        
        size_t prev = remove_sub(index, NO_REF);

        size_t child = orphan(index, prev); 
        
        reunite(child);
    } 
    
    template <typename Key>
    size_t BinomialQueue<Key>::remove_sub(size_t index, size_t prev)
    { 
        size_t parent = m_nodes[index].m_rchild;
        while ((parent & PARENT_REF) == 0)
        {
            parent = m_nodes[parent].m_rchild;
        } 
                
        if (parent != NO_REF)
        {
            parent &= ~PARENT_REF;

            prev = remove_sub(parent, prev);

            size_t child = m_nodes[parent].m_lchild;
            
            while (child != index)
            {
                size_t next = m_nodes[child].m_rchild;
                m_nodes[child].m_rchild = prev;
                prev = child;
                child = next;
            }
            
            m_nodes[parent].m_lchild = (m_nodes[index].m_rchild & PARENT_REF) == 0 ? m_nodes[index].m_rchild : NO_REF;
            m_nodes[parent].m_rchild = prev;
            prev = parent;
        }

        return prev;
    }
}

#endif
