/*
 * GUTS - Generic Utilities 
 * Copyright (c) 2006-2010 Gino van den Bergen, DTECTA 
 * http://www.dtecta.com
 *
 * Permission to use, copy, modify, distribute and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear
 * in supporting documentation.  Gino van den Bergen makes no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied warranty.
 */

//! @file HashTable.hpp guts: Hash table.

#ifndef GUTS_HASHTABLE_HPP
#define GUTS_HASHTABLE_HPP

#include "Allocator.hpp"
#include "Vector.hpp"
#include "Invariant.hpp"

#include <string.h>

// Updating the hash table the "smart" way is actually slower
#define WASTE_CYCLES 0

namespace guts
{
    template <typename Value>
    struct Hash
    {
        size_t operator()(const Value& value) const
        {
            return hash(value); 
        } 
    };

    template <typename Value, typename HashFunc = Hash<Value>, typename Alloc = Allocator<Value> >
    class HashTable
    {       
    public:   
        HashTable(size_t nbuckets)
            : m_nbuckets(nbuckets)
        {
            GUARD_INVARIANT_AT_EXIT(this);

            m_bucketCounts = ALLOCATE_ARRAY(size_t, nbuckets);

            memset(m_bucketCounts, 0, sizeof(size_t) * nbuckets);
        }  

        explicit HashTable(size_t nbuckets, const HashFunc& hashFunc)
            : m_nbuckets(nbuckets)
            , m_hashFunc(hashFunc)
        {
            GUARD_INVARIANT_AT_EXIT(this);  

            m_bucketCounts = ALLOCATE_ARRAY(size_t, nbuckets);

            memset(m_bucketCounts, 0, sizeof(size_t) * nbuckets);
        }

        ~HashTable()
        {
            DEALLOCATE(m_bucketCounts);
        }

        bool empty() const
        {
            return m_entries.empty();
        }

        void insert(const Value& value)
        {
            GUARD_INVARIANT(this);

            size_t bucket = m_hashFunc(value) % m_nbuckets;

#if WASTE_CYCLES
            m_entries.push_back(Value());
            size_t last = m_bucketCounts[m_nbuckets-1]++;
            for (size_t i = m_nbuckets-1; i != bucket; --i)
            {
                size_t first = m_bucketCounts[i-1]++;
            
                if (first < last)
                {
                    m_entries[last] = m_entries[first];
                }
                last = first; 
            }  

            m_entries[last] = value;
#else
            m_entries.insert(m_entries.begin() + m_bucketCounts[bucket], value); 
            for (size_t i = bucket; i != m_nbuckets; ++i)
            {
                ++m_bucketCounts[i];
            }
#endif

        }

        const Value* retrieve(const Value& value) const
        {
            size_t bucket = m_hashFunc(value) % m_nbuckets;

            typename Heap::const_iterator it = m_entries.begin();
            if (bucket != 0)
            {
                it += m_bucketCounts[bucket-1];
            }
            typename Heap::const_iterator last = m_entries.begin() + m_bucketCounts[bucket]; 
            while (it != last && !(*it == value))
            {
                ++it;
            }
            
            return it != last ? &(*it) : 0;
        }

        void remove(const Value& value)
        {
            GUARD_INVARIANT(this);

            size_t bucket = m_hashFunc(value) % m_nbuckets; 

#if WASTE_CYCLES
            size_t i = bucket != 0 ? m_bucketCounts[bucket-1] : 0, last = m_bucketCounts[bucket]; 
            while (i != last && m_entries[i] != value)
            {
                ++i;
            }
            
            ASSERT(i != last);

            m_entries[i] = m_entries[last-1];
            --m_bucketCounts[bucket];
            
            for (size_t i = bucket+1; i != m_nbuckets; ++i)
            {
                size_t first = m_bucketCounts[i-1];
                size_t last = --m_bucketCounts[i];
                if (first < last)
                {
                    m_entries[first] = m_entries[last];
                }
            }  

            m_entries.pop_back();
#else
            typename Heap::iterator it = m_entries.begin();
            if (bucket != 0)
            {
                it += m_bucketCounts[bucket-1];
            }

            typename Heap::iterator last = m_entries.begin() + m_bucketCounts[bucket]; 
            while (it != last && !(*it == value))
            {
                ++it;
            }
            
           ASSERT(it != last);

           m_entries.erase(it); 
           for (size_t i = bucket; i != m_nbuckets; ++i)
           {
               --m_bucketCounts[i];
           }         
#endif
        }

    protected:
        typedef Value Entry; // second is a pointer to an index referring to this entry. So, if 
                                              // an entry is moved, we can update the reference.
        typedef typename VectorFix<Entry, Alloc>::RT Heap;      // random access container  

        Heap m_entries;
        size_t m_nbuckets;
        size_t* m_bucketCounts;
        HashFunc m_hashFunc;         

        INVARIANT
        {
            typename Heap::const_iterator first = m_entries.begin();
            for (size_t i = 0; i != m_nbuckets; ++i)
            {
                 typename Heap::const_iterator last = m_entries.begin() + m_bucketCounts[i];
                 ASSERT(first <= last && last <= m_entries.end());
                 while (first != last)
                 {
                     ASSERT((m_hashFunc(*first) % m_nbuckets) == i);
                     ++first; 
                 }
            }
            ASSERT(m_bucketCounts[m_nbuckets-1] == m_entries.size());
        }
    };
}


#endif
