/*
 * GUTS - Generic Utilities 
 * Copyright (c) 2006-2010 Gino van den Bergen, DTECTA 
 * http://www.dtecta.com
 *
 * Permission to use, copy, modify, distribute and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear
 * in supporting documentation.  Gino van den Bergen makes no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied warranty.
 */

#ifndef GUTS_REFPTR_HPP
#define GUTS_REFPTR_HPP

#include "Memory.hpp"

#include <tbb/atomic.h>

namespace guts
{
    class RefCounted
    {
    public:
        RefCounted()
        {
            m_refCount = 0;
        }

        void increase()
        { 
            m_refCount++; 
        }

        void decrease()
        {
            ASSERT(m_refCount != 0);
            if (--m_refCount == 0)
            {
                this->~RefCounted();
                DEALLOCATE(this);
            }
        }

        uint32_t refCount() const { return m_refCount; }

    protected:
         virtual ~RefCounted() {}

    private:
        tbb::atomic<uint32_t> m_refCount; 
    };

    template <typename T>
    class RefPtr
    {
    public:
        RefPtr(T* ptr = 0)
        {
            acquire(ptr);
        } 

        RefPtr(const RefPtr<T>& rhs)
        {
            acquire(rhs.get());
        }

        template <typename U>
        RefPtr(const RefPtr<U>& rhs)
        {
            acquire(rhs.get());
        }

        ~RefPtr()
        {
            release();  
        }

        RefPtr<T>& operator=(T* ptr)
        {
            if (m_ptr != ptr)
            {
                release();
                acquire(ptr);
            }
            return *this;
        }

        RefPtr<T>& operator=(const RefPtr<T>& rhs)
        {
            return *this = rhs.get();
        }

        template <typename U>
        RefPtr<T>& operator=(const RefPtr<U>& rhs)
        {
            return *this = rhs.get();
        }

        T& operator*() const 
        {
            ASSERT(m_ptr != 0 && m_ptr->refCount() != 0); 
            return *m_ptr; 
        }

        T* operator->() const 
        {
            ASSERT(m_ptr != 0 && m_ptr->refCount() != 0); 
            return m_ptr; 
        }
        T* get() const { return m_ptr; }
        bool operator!() const { return m_ptr == 0; }

    private:
        void acquire(T* ptr)
        {
            m_ptr = ptr;
            if (m_ptr != 0)
            {
                m_ptr->increase();
            }
        }
        
       void release()
        {
            if (m_ptr != 0)
            {
                m_ptr->decrease();
            }
            // m_ptr = 0; is not necessary, since release is not called directly.
        }

        T* m_ptr;
    }; 

    template <typename T>
    FORCEINLINE
    bool operator==(const RefPtr<T>& lhs, int)
    {
        return lhs.get() == 0; 
    }

    template <typename T>
    FORCEINLINE
    bool operator==(int, const RefPtr<T>& rhs)
    {
        return 0 == rhs.get(); 
    }

    template <typename T, typename U>
    FORCEINLINE
    bool operator==(const RefPtr<T>& lhs, const U* rhs)
    {
        return lhs.get() == rhs; 
    }

    template <typename T, typename U>
    FORCEINLINE
    bool operator==(const T* lhs, const RefPtr<U>& rhs)
    {
        return lhs == rhs.get();
    }

    template <typename T, typename U>
    FORCEINLINE
    bool operator==(const RefPtr<T>& lhs, const RefPtr<U>& rhs)
    {
        return lhs.get() == rhs.get(); 
    }

    template <typename T>
    FORCEINLINE
    bool operator!=(const RefPtr<T>& lhs, int)
    {
        return lhs.get() != 0; 
    }

    template <typename T>
    FORCEINLINE
    bool operator!=(int, const RefPtr<T>& rhs)
    {
        return 0 != rhs.get(); 
    }

    template <typename T, typename U>
    FORCEINLINE
    bool operator!=(const RefPtr<T>& lhs, const U* rhs)
    {
        return lhs.get() != rhs; 
    }

    template <typename T, typename U>
    FORCEINLINE
    bool operator!=(const T* lhs, const RefPtr<U>& rhs)
    {
        return lhs != rhs.get();
    }

    template <typename T, typename U>
    FORCEINLINE
    bool operator!=(const RefPtr<T>& lhs, const RefPtr<U>& rhs)
    {
        return lhs.get() != rhs.get(); 
    }
}

#endif
