/*
 * GUTS - Generic Utilities 
 * Copyright (c) 2006-2010 Gino van den Bergen, DTECTA 
 * http://www.dtecta.com
 *
 * Permission to use, copy, modify, distribute and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear
 * in supporting documentation.  Gino van den Bergen makes no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied warranty.
 */

#ifndef GUTS_SPATIALHASH_HPP
#define GUTS_SPATIALHASH_HPP


#include "HashTable.hpp"

namespace guts
{
    static const size_t g_primes[] = { 73856093, 19349663, 83492791 };  
    
    template <typename Value>
    struct HashFunc
    { 
        typedef typename Value::ScalarType ScalarType;
        enum { DIM = Value::DIM };
            
        HashFunc(typename Value::ScalarType cellSize)
            : m_cellSize(cellSize)
        {}

        size_t operator()(const typename Value::ScalarType* value) const;
 
        ScalarType m_cellSize; 
    };  

    template <typename Value>
    size_t HashFunc<Value>::operator()(const typename Value::ScalarType* v) const 
    {
        size_t result = 0; 
        for (int i = 0; i != DIM; ++i)
        {
            result ^= size_t(v[i] / m_cellSize + ScalarType(0.5)) * g_primes[i];
        }

        return result;
    } 
    
    template <typename Value>
    class SpatialHash
        : public HashTable<Value, HashFunc<Value> >
    {
    public:
        SpatialHash(size_t nbuckets, typename Value::ScalarType cellSize);
                
        template <typename Function>
        void query(const typename Value::ScalarType* v, typename Value::ScalarType range, Function& callback) const;
    };   

    template <typename Value>
    SpatialHash<Value>::SpatialHash(size_t nbuckets, typename Value::ScalarType cellSize)
        : HashTable<Value, HashFunc<Value> >(nbuckets, HashFunc<Value>(cellSize))
    {} 
    
     
  
    
    template <typename Value>
    template <typename Function>
    void SpatialHash<Value>::query(const typename Value::ScalarType* v, typename Value::ScalarType range, Function& callback) const
    {
        ASSERT(range * typename Value::ScalarType(2) <= this->m_hashFunc.m_cellSize);
        static const unsigned int MaxBuckets = 1<<Value::DIM;
        size_t buckets[MaxBuckets];
        int bucketCount = 0;
        for (unsigned int bits = 0; bits != MaxBuckets; ++bits)
        {
            typename Value::ScalarType pos[Value::DIM];
            for (int i = 0; i != Value::DIM; ++i)
            {
                pos[i] = v[i] + ((bits & (1 << i)) ? -range : range);  
            }
            
            size_t bucket = m_hashFunc(pos) % this->m_nbuckets;

            int k = 0;
            while (k != bucketCount && buckets[k] != bucket)
            {
                ++k;
            }
            if (k == bucketCount)
            {
                buckets[bucketCount++] = bucket; 
            }
        }
                        
        for (int i = 0; i != bucketCount; ++i)
        {  
            size_t bucket = buckets[i];

            typename HashTable<Value, HashFunc<Value> >::Heap::const_iterator it = this->m_entries.begin();
            if (bucket != 0)
            {
                it += this->m_bucketCounts[bucket-1];
            }
            typename HashTable<Value, HashFunc<Value> >::Heap::const_iterator last = this->m_entries.begin() + this->m_bucketCounts[bucket]; 
            while (it != last)
            { 
                callback(*it);
                ++it;
            }
        }    
    }
}

#endif
